@@ -1,27 +1,7 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
-/*
- * Copyright © 2016 Endless Mobile, Inc.
- *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to
- * deal in the Software without restriction, including without limitation the
- * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
- * sell copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
- * IN THE SOFTWARE.
- *
- * Authored by: Philip Chimento <philip@endlessm.com>
- */
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2016 Endless Mobile, Inc.
+// SPDX-FileContributor: Authored by: Philip Chimento <philip@endlessm.com>
 
 #ifndef GJS_JSAPI_UTIL_ARGS_H_
 #define GJS_JSAPI_UTIL_ARGS_H_
@@ -37,7 +17,9 @@
 
 #include <js/CallArgs.h>
 #include <js/Conversions.h>
+#include <js/RootingAPI.h>
 #include <js/TypeDecls.h>
+#include <js/Utility.h>  // for UniqueChars
 #include <jsapi.h>  // for JS_ClearPendingException
 
 #include "cjs/jsapi-util.h"
@@ -69,8 +51,6 @@ static inline void assign(JSContext*, char c, bool nullable,
     *ref = value.toBoolean();
 }
 
-/* This preserves the previous behaviour of gjs_parse_args(), but maybe we want
- * to box primitive types instead of throwing? */
 GJS_ALWAYS_INLINE
 static inline void assign(JSContext*, char c, bool nullable,
                           JS::HandleValue value, JS::MutableHandleObject ref) {
@@ -118,6 +98,21 @@ assign(JSContext      *cx,
         throw g_strdup("Couldn't convert to filename");
 }
 
+GJS_ALWAYS_INLINE
+static inline void assign(JSContext*, char c, bool nullable,
+                          JS::HandleValue value, JS::MutableHandleString ref) {
+    if (c != 'S')
+        throw g_strdup_printf("Wrong type for %c, got JS::MutableHandleString",
+                              c);
+    if (nullable && value.isNull()) {
+        ref.set(nullptr);
+        return;
+    }
+    if (!value.isString())
+        throw g_strdup("Not a string");
+    ref.set(value.toString());
+}
+
 GJS_ALWAYS_INLINE
 static inline void
 assign(JSContext      *cx,
@@ -204,20 +199,16 @@ GJS_ALWAYS_INLINE static inline void assign(JSContext* cx, char c,
     assign(cx, c, nullable, value, (int *)ref);
 }
 
-/* Force JS::RootedObject * to be converted to JS::MutableHandleObject,
- * see overload in jsapi-util-args.cpp */
-template <typename T, typename std::enable_if_t<
-                          !std::is_same_v<T, JS::RootedObject*>, int> = 0>
+template <typename T>
 static inline void free_if_necessary(T param_ref [[maybe_unused]]) {}
 
-GJS_ALWAYS_INLINE
-static inline void
-free_if_necessary(JS::MutableHandleObject param_ref)
-{
-    /* This is not exactly right, since before we consumed a JS::ObjectValue
-     * there may have been something different inside the handle. But it has
-     * already been clobbered at this point anyhow */
-    param_ref.set(nullptr);
+template <typename T>
+GJS_ALWAYS_INLINE static inline void free_if_necessary(
+    JS::Rooted<T>* param_ref) {
+    // This is not exactly right, since before we consumed a JS::Value there may
+    // have been something different inside the handle. But it has already been
+    // clobbered at this point anyhow.
+    JS::MutableHandle<T>(param_ref).set(nullptr);
 }
 
 template <typename T>
@@ -272,7 +263,7 @@ GJS_JSAPI_RETURN_CONVENTION static bool parse_call_args_helper(
     bool retval = parse_call_args_helper(cx, function_name, args, fmt_required,
                                          fmt_optional, ++param_ix, params...);
 
-    /* We still own the strings in the error case, free any we converted */
+    // We still own JSString/JSObject in the error case, free any we converted
     if (!retval)
         free_if_necessary(param_ref);
     return retval;
@@ -305,8 +296,8 @@ GJS_JSAPI_RETURN_CONVENTION [[maybe_unused]] static bool gjs_parse_call_args(
  * gjs_parse_call_args:
  * @context:
  * @function_name: The name of the function being called
- * @format: Printf-like format specifier containing the expected arguments
  * @args: #JS::CallArgs from #JSNative function
+ * @format: Printf-like format specifier containing the expected arguments
  * @params: for each character in @format, a pair of const char * which is the
  * name of the argument, and a location to store the value. The type of
  * location argument depends on the format character, as described below.
@@ -320,6 +311,7 @@ GJS_JSAPI_RETURN_CONVENTION [[maybe_unused]] static bool gjs_parse_call_args(
  * s: A string, converted into UTF-8 (pass a JS::UniqueChars*)
  * F: A string, converted into "filename encoding" (i.e. active locale) (pass
  *   a GjsAutoChar *)
+ * S: A string, no conversion (pass a JS::MutableHandleString)
  * i: A number, will be converted to a 32-bit int (pass an int32_t * or a
  *   pointer to an enum type)
  * u: A number, converted into a 32-bit unsigned int (pass a uint32_t *)
@@ -334,9 +326,9 @@ GJS_JSAPI_RETURN_CONVENTION [[maybe_unused]] static bool gjs_parse_call_args(
  * after a '|' when not specified, do not cause any changes in the C
  * value location.
  *
- * A prefix character '?' in front of 's', 'F', or 'o' means that the next value
- * may be null. For 's' or 'F' a null pointer is returned, for 'o' the handle is
- * set to null.
+ * A prefix character '?' in front of 's', 'F', 'S', or 'o' means that the next
+ * value may be null. For 's' or 'F' a null pointer is returned, for 'S' or 'o'
+ * the handle is set to null.
  */
 template <typename... Args>
 GJS_JSAPI_RETURN_CONVENTION static bool gjs_parse_call_args(
