@@ -1,43 +1,27 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
-/*
- * Copyright (c) 2008  litl, LLC
- *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to
- * deal in the Software without restriction, including without limitation the
- * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
- * sell copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
- * IN THE SOFTWARE.
- */
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2008 litl, LLC
+// SPDX-FileCopyrightText: 2018-2020  Canonical, Ltd
 
 #ifndef GJS_JSAPI_UTIL_H_
 #define GJS_JSAPI_UTIL_H_
 
 #include <config.h>
 
-#include <stddef.h>  // for size_t
 #include <stdint.h>
+#include <stdlib.h>     // for free
 #include <sys/types.h>  // for ssize_t
 
-#include <memory>  // for unique_ptr
 #include <string>  // for string, u16string
+#include <type_traits>  // for enable_if_t, add_pointer_t, add_const_t
+#include <utility>      // IWYU pragma: keep
 #include <vector>
 
 #include <girepository.h>
 #include <glib-object.h>
 #include <glib.h>
 
+#include <js/GCAPI.h>
 #include <js/GCPolicyAPI.h>  // for IgnoreGCPolicy
 #include <js/Id.h>
 #include <js/TypeDecls.h>
@@ -49,40 +33,174 @@
 class JSErrorReport;
 namespace JS {
 class CallArgs;
-}
+
+struct Dummy {};
+using GTypeNotUint64 =
+    std::conditional_t<!std::is_same_v<GType, uint64_t>, GType, Dummy>;
+
+// The GC sweep method should ignore FundamentalTable and GTypeTable's key types
+// Forward declarations
+template <>
+struct GCPolicy<void*> : public IgnoreGCPolicy<void*> {};
+// We need GCPolicy<GType> for GTypeTable. SpiderMonkey already defines
+// GCPolicy<uint64_t> which is equal to GType on some systems; for others we
+// need to define it. (macOS's uint64_t is unsigned long long, which is a
+// different type from unsigned long, even if they are the same width)
+template <>
+struct GCPolicy<GTypeNotUint64> : public IgnoreGCPolicy<GTypeNotUint64> {};
+}  // namespace JS
 
 struct GjsAutoTakeOwnership {};
 
-template <typename T, typename F,
-          void (*free_func)(F*) = nullptr, F* (*ref_func)(F*) = nullptr>
-struct GjsAutoPointer : std::unique_ptr<T, decltype(free_func)> {
-    GjsAutoPointer(T* ptr = nullptr)  // NOLINT(runtime/explicit)
-        : GjsAutoPointer::unique_ptr(ptr, *free_func) {}
-    GjsAutoPointer(T* ptr, const GjsAutoTakeOwnership&)
-        : GjsAutoPointer(nullptr) {
-        // FIXME: should use if constexpr (...), but that doesn't work with
-        // ubsan, which generates a null pointer check making it not a constexpr
-        // anymore: https://gcc.gnu.org/bugzilla/show_bug.cgi?id=71962 - Also a
-        // bogus warning, https://gcc.gnu.org/bugzilla/show_bug.cgi?id=94554
-        auto ref = ref_func;
-        this->reset(ptr && ref ? reinterpret_cast<T*>(ref(ptr)) : ptr);
+template <typename F = void>
+using GjsAutoPointerRefFunction = F* (*)(F*);
+
+template <typename F = void>
+using GjsAutoPointerFreeFunction = void (*)(F*);
+
+template <typename T, typename F = void,
+          GjsAutoPointerFreeFunction<F> free_func = free,
+          GjsAutoPointerRefFunction<F> ref_func = nullptr>
+struct GjsAutoPointer {
+    using Tp =
+        std::conditional_t<std::is_array_v<T>, std::remove_extent_t<T>, T>;
+    using Ptr = std::add_pointer_t<Tp>;
+    using ConstPtr = std::add_pointer_t<std::add_const_t<Tp>>;
+
+ private:
+    template <typename FunctionType, FunctionType function>
+    static constexpr bool has_function() {
+        using NullType = std::integral_constant<FunctionType, nullptr>;
+        using ActualType = std::integral_constant<FunctionType, function>;
+
+        return !std::is_same_v<ActualType, NullType>;
     }
 
-    operator T*() const { return this->get(); }
-    T& operator[](size_t i) const { return static_cast<T*>(*this)[i]; }
+ public:
+    static constexpr bool has_free_function() {
+        return has_function<GjsAutoPointerFreeFunction<F>, free_func>();
+    }
 
-    [[nodiscard]] T* copy() const {
-        return reinterpret_cast<T*>(ref_func(this->get()));
+    static constexpr bool has_ref_function() {
+        return has_function<GjsAutoPointerRefFunction<F>, ref_func>();
+    }
+
+    constexpr GjsAutoPointer(Ptr ptr = nullptr)  // NOLINT(runtime/explicit)
+        : m_ptr(ptr) {}
+    template <typename U, typename = std::enable_if_t<std::is_same_v<U, Tp> &&
+                                                      std::is_array_v<T>>>
+    explicit constexpr GjsAutoPointer(U ptr[]) : m_ptr(ptr) {}
+
+    constexpr GjsAutoPointer(Ptr ptr, const GjsAutoTakeOwnership&)
+        : GjsAutoPointer(ptr) {
+        m_ptr = copy();
+    }
+    constexpr GjsAutoPointer(ConstPtr ptr, const GjsAutoTakeOwnership& o)
+        : GjsAutoPointer(const_cast<Ptr>(ptr), o) {}
+    constexpr GjsAutoPointer(GjsAutoPointer&& other) : GjsAutoPointer() {
+        this->swap(other);
+    }
+    constexpr GjsAutoPointer(GjsAutoPointer const& other) : GjsAutoPointer() {
+        *this = other;
+    }
+
+    constexpr GjsAutoPointer& operator=(Ptr ptr) {
+        reset(ptr);
+        return *this;
+    }
+
+    GjsAutoPointer& operator=(GjsAutoPointer&& other) {
+        this->swap(other);
+        return *this;
+    }
+
+    GjsAutoPointer& operator=(GjsAutoPointer const& other) {
+        GjsAutoPointer dup(other.get(), GjsAutoTakeOwnership());
+        this->swap(dup);
+        return *this;
+    }
+
+    template <typename U = T>
+    constexpr std::enable_if_t<!std::is_array_v<U>, Ptr> operator->() {
+        return m_ptr;
+    }
+
+    template <typename U = T>
+    constexpr std::enable_if_t<!std::is_array_v<U>, ConstPtr> operator->()
+        const {
+        return m_ptr;
+    }
+
+    constexpr Tp operator*() const { return *m_ptr; }
+    constexpr operator Ptr() { return m_ptr; }
+    constexpr operator Ptr() const { return m_ptr; }
+    constexpr operator ConstPtr() const { return m_ptr; }
+    constexpr operator bool() const { return m_ptr != nullptr; }
+
+    constexpr Ptr get() const { return m_ptr; }
+    constexpr Ptr* out() { return &m_ptr; }
+
+    constexpr Ptr release() {
+        auto* ptr = m_ptr;
+        m_ptr = nullptr;
+        return ptr;
+    }
+
+    constexpr void reset(Ptr ptr = nullptr) {
+        Ptr old_ptr = m_ptr;
+        m_ptr = ptr;
+
+        if constexpr (has_free_function()) {
+            if (old_ptr) {
+                if constexpr (std::is_array_v<T>)
+                    free_func(reinterpret_cast<T*>(old_ptr));
+                else
+                    free_func(old_ptr);
+            }
+        }
+    }
+
+    constexpr void swap(GjsAutoPointer& other) {
+        std::swap(this->m_ptr, other.m_ptr);
+    }
+
+    /* constexpr */ ~GjsAutoPointer() {  // one day, with -std=c++2a
+        reset();
+    }
+
+    template <typename U = T>
+    [[nodiscard]] constexpr std::enable_if_t<!std::is_array_v<U>, Ptr> copy()
+        const {
+        static_assert(has_ref_function(), "No ref function provided");
+        return m_ptr ? reinterpret_cast<Ptr>(ref_func(m_ptr)) : nullptr;
     }
 
     template <typename C>
-    [[nodiscard]] C* as() const {
-        return const_cast<C*>(reinterpret_cast<const C*>(this->get()));
+    [[nodiscard]] constexpr C* as() const {
+        return const_cast<C*>(reinterpret_cast<const C*>(m_ptr));
     }
+
+ private:
+    Ptr m_ptr;
 };
 
+template <typename T, GjsAutoPointerFreeFunction<T> free_func = free,
+          GjsAutoPointerRefFunction<T> ref_func = nullptr>
+struct GjsAutoPointerSimple : GjsAutoPointer<T, T, free_func, ref_func> {
+    using GjsAutoPointer<T, T, free_func, ref_func>::GjsAutoPointer;
+};
+
+template <typename T, typename F = void,
+          GjsAutoPointerFreeFunction<F> free_func,
+          GjsAutoPointerRefFunction<F> ref_func>
+constexpr bool operator==(
+    GjsAutoPointer<T, F, free_func, ref_func> const& lhs,
+    GjsAutoPointer<T, F, free_func, ref_func> const& rhs) {
+    return lhs.get() == rhs.get();
+}
+
 template <typename T>
-using GjsAutoFree = GjsAutoPointer<T, void, g_free>;
+using GjsAutoFree = GjsAutoPointer<T>;
 
 struct GjsAutoCharFuncs {
     static char* dup(char* str) { return g_strdup(str); }
@@ -91,11 +209,33 @@ struct GjsAutoCharFuncs {
 using GjsAutoChar =
     GjsAutoPointer<char, char, GjsAutoCharFuncs::free, GjsAutoCharFuncs::dup>;
 
-using GjsAutoStrv = GjsAutoPointer<char*, char*, g_strfreev>;
+using GjsAutoChar16 = GjsAutoPointer<uint16_t, void, &g_free>;
+
+struct GjsAutoErrorFuncs {
+    static GError* error_copy(GError* error) { return g_error_copy(error); }
+};
+using GjsAutoError =
+    GjsAutoPointer<GError, GError, g_error_free, GjsAutoErrorFuncs::error_copy>;
+
+using GjsAutoStrv = GjsAutoPointer<char*, char*, g_strfreev, g_strdupv>;
 
 template <typename T>
 using GjsAutoUnref = GjsAutoPointer<T, void, g_object_unref, g_object_ref>;
 
+using GjsAutoGVariant =
+    GjsAutoPointer<GVariant, GVariant, g_variant_unref, g_variant_ref>;
+
+template <typename V, typename T>
+constexpr void GjsAutoPointerDeleter(T v) {
+    if constexpr (std::is_array_v<V>)
+        delete[] reinterpret_cast<std::remove_extent_t<V>*>(v);
+    else
+        delete v;
+}
+
+template <typename T>
+using GjsAutoCppPointer = GjsAutoPointer<T, T, GjsAutoPointerDeleter<T>>;
+
 template <typename T = GTypeClass>
 struct GjsAutoTypeClass : GjsAutoPointer<T, void, &g_type_class_unref> {
     GjsAutoTypeClass(gpointer ptr = nullptr)  // NOLINT(runtime/explicit)
@@ -110,8 +250,7 @@ struct GjsAutoTypeClass : GjsAutoPointer<T, void, &g_type_class_unref> {
 // returns GIFunctionInfo*,) use one of the derived classes below.
 struct GjsAutoBaseInfo : GjsAutoPointer<GIBaseInfo, GIBaseInfo,
                                         g_base_info_unref, g_base_info_ref> {
-    GjsAutoBaseInfo(GIBaseInfo* ptr = nullptr)  // NOLINT(runtime/explicit)
-        : GjsAutoPointer(ptr) {}
+    using GjsAutoPointer::GjsAutoPointer;
 
     [[nodiscard]] const char* name() const {
         return g_base_info_get_name(*this);
@@ -128,6 +267,8 @@ struct GjsAutoBaseInfo : GjsAutoPointer<GIBaseInfo, GIBaseInfo,
 // the info is either of a certain type or null.
 template <GIInfoType TAG>
 struct GjsAutoInfo : GjsAutoBaseInfo {
+    using GjsAutoBaseInfo::GjsAutoBaseInfo;
+
     // Normally one-argument constructors should be explicit, but we are trying
     // to conform to the interface of std::unique_ptr here.
     GjsAutoInfo(GIBaseInfo* ptr = nullptr)  // NOLINT(runtime/explicit)
@@ -164,6 +305,8 @@ using GjsAutoVFuncInfo = GjsAutoInfo<GI_INFO_TYPE_VFUNC>;
 // GICallableInfo can be one of several tags, so we have to have a separate
 // class, and use GI_IS_CALLABLE_INFO() to validate.
 struct GjsAutoCallableInfo : GjsAutoBaseInfo {
+    using GjsAutoBaseInfo::GjsAutoBaseInfo;
+
     GjsAutoCallableInfo(GIBaseInfo* ptr = nullptr)  // NOLINT(runtime/explicit)
         : GjsAutoBaseInfo(ptr) {
         validate();
@@ -176,11 +319,56 @@ struct GjsAutoCallableInfo : GjsAutoBaseInfo {
 
  private:
     void validate() const {
-        if (GIBaseInfo* base = *this)
-            g_assert(GI_IS_CALLABLE_INFO(base));
+        if (*this)
+            g_assert(GI_IS_CALLABLE_INFO(get()));
     }
 };
 
+template <typename T>
+struct GjsSmartPointer : GjsAutoPointer<T> {
+    using GjsAutoPointer<T>::GjsAutoPointer;
+};
+
+template <>
+struct GjsSmartPointer<char*> : GjsAutoStrv {
+    using GjsAutoStrv::GjsAutoPointer;
+};
+
+template <>
+struct GjsSmartPointer<GStrv> : GjsAutoStrv {
+    using GjsAutoStrv::GjsAutoPointer;
+};
+
+template <>
+struct GjsSmartPointer<GObject> : GjsAutoUnref<GObject> {
+    using GjsAutoUnref<GObject>::GjsAutoUnref;
+};
+
+template <>
+struct GjsSmartPointer<GIBaseInfo> : GjsAutoBaseInfo {
+    using GjsAutoBaseInfo::GjsAutoBaseInfo;
+};
+
+template <>
+struct GjsSmartPointer<GError> : GjsAutoError {
+    using GjsAutoError::GjsAutoError;
+};
+
+template <>
+struct GjsSmartPointer<GVariant> : GjsAutoGVariant {
+    using GjsAutoGVariant::GjsAutoPointer;
+};
+
+template <>
+struct GjsSmartPointer<GList> : GjsAutoPointer<GList, GList, g_list_free> {
+    using GjsAutoPointer::GjsAutoPointer;
+};
+
+template <>
+struct GjsSmartPointer<GSList> : GjsAutoPointer<GSList, GSList, g_slist_free> {
+    using GjsAutoPointer::GjsAutoPointer;
+};
+
 /* For use of GjsAutoInfo<TAG> in GC hash maps */
 namespace JS {
 template <GIInfoType TAG>
@@ -223,10 +411,12 @@ struct GCPolicy<GjsAutoParam> : public IgnoreGCPolicy<GjsAutoParam> {};
 
 [[nodiscard]] JSObject* gjs_get_import_global(JSContext* cx);
 
+[[nodiscard]] JSObject* gjs_get_internal_global(JSContext* cx);
+
 void gjs_throw_constructor_error             (JSContext       *context);
 
-void gjs_throw_abstract_constructor_error(JSContext    *context,
-                                          JS::CallArgs& args);
+void gjs_throw_abstract_constructor_error(JSContext* cx,
+                                          const JS::CallArgs& args);
 
 GJS_JSAPI_RETURN_CONVENTION
 JSObject* gjs_build_string_array(JSContext* cx,
@@ -238,14 +428,12 @@ JSObject* gjs_define_string_array(JSContext* cx, JS::HandleObject obj,
                                   const std::vector<std::string>& strings,
                                   unsigned attrs);
 
-void        gjs_throw                        (JSContext       *context,
-                                              const char      *format,
-                                              ...)  G_GNUC_PRINTF (2, 3);
-void        gjs_throw_custom                 (JSContext       *context,
-                                              JSProtoKey       error_kind,
-                                              const char      *error_name,
-                                              const char      *format,
-                                              ...)  G_GNUC_PRINTF (4, 5);
+[[gnu::format(printf, 2, 3)]] void gjs_throw(JSContext* cx, const char* format,
+                                             ...);
+[[gnu::format(printf, 4, 5)]] void gjs_throw_custom(JSContext* cx,
+                                                    JSProtoKey error_kind,
+                                                    const char* error_name,
+                                                    const char* format, ...);
 void        gjs_throw_literal                (JSContext       *context,
                                               const char      *string);
 bool gjs_throw_gerror_message(JSContext* cx, GError* error);
@@ -257,14 +445,22 @@ bool gjs_log_exception_uncaught(JSContext* cx);
 bool gjs_log_exception_full(JSContext* cx, JS::HandleValue exc,
                             JS::HandleString message, GLogLevelFlags level);
 
-[[nodiscard]] char* gjs_value_debug_string(JSContext* cx,
-                                           JS::HandleValue value);
+[[nodiscard]] std::string gjs_value_debug_string(JSContext* cx,
+                                                 JS::HandleValue value);
 
 void gjs_warning_reporter(JSContext*, JSErrorReport* report);
 
 GJS_JSAPI_RETURN_CONVENTION
 JS::UniqueChars gjs_string_to_utf8(JSContext* cx, const JS::Value string_val);
 GJS_JSAPI_RETURN_CONVENTION
+bool gjs_string_to_utf8_n(JSContext* cx, JS::HandleString str, JS::UniqueChars* output,
+                          size_t* output_len);
+GJS_JSAPI_RETURN_CONVENTION
+JSString* gjs_lossy_string_from_utf8(JSContext* cx, const char* utf8_string);
+GJS_JSAPI_RETURN_CONVENTION
+JSString* gjs_lossy_string_from_utf8_n(JSContext* cx, const char* utf8_string,
+                                       size_t len);
+GJS_JSAPI_RETURN_CONVENTION
 bool gjs_string_from_utf8(JSContext             *context,
                           const char            *utf8_string,
                           JS::MutableHandleValue value_p);
@@ -318,9 +514,6 @@ bool        gjs_unichar_from_string          (JSContext       *context,
 void gjs_maybe_gc (JSContext *context);
 void gjs_gc_if_needed(JSContext *cx);
 
-[[nodiscard]] std::u16string gjs_utf8_script_to_utf16(const char* script,
-                                                      ssize_t len);
-
 GJS_JSAPI_RETURN_CONVENTION
 GjsAutoChar gjs_format_stack_trace(JSContext       *cx,
                                    JS::HandleObject saved_frame);
@@ -375,10 +568,41 @@ bool gjs_object_require_converted_property(JSContext       *context,
 [[nodiscard]] std::string gjs_debug_value(JS::Value v);
 [[nodiscard]] std::string gjs_debug_id(jsid id);
 
-[[nodiscard]] char* gjs_hyphen_to_underscore(const char* str);
+[[nodiscard]] GjsAutoChar gjs_hyphen_to_underscore(const char* str);
+[[nodiscard]] GjsAutoChar gjs_hyphen_to_camel(const char* str);
 
 #if defined(G_OS_WIN32) && (defined(_MSC_VER) && (_MSC_VER >= 1900))
 [[nodiscard]] std::wstring gjs_win32_vc140_utf8_to_utf16(const char* str);
 #endif
 
+// Custom GC reasons; SpiderMonkey includes a bunch of "Firefox reasons" which
+// don't apply when embedding the JS engine, so we repurpose them for our own
+// reasons.
+
+// clang-format off
+#define FOREACH_GC_REASON(macro)  \
+    macro(LINUX_RSS_TRIGGER, 0)   \
+    macro(GJS_CONTEXT_DISPOSE, 1) \
+    macro(BIG_HAMMER, 2)          \
+    macro(GJS_API_CALL, 3)
+// clang-format on
+
+namespace Gjs {
+
+struct GCReason {
+#define DEFINE_GC_REASON(name, ix)                     \
+    static constexpr JS::GCReason name = JS::GCReason( \
+        static_cast<int>(JS::GCReason::FIRST_FIREFOX_REASON) + ix);
+FOREACH_GC_REASON(DEFINE_GC_REASON);
+#undef DEFINE_GC_REASON
+
+#define COUNT_GC_REASON(name, ix) +1
+static constexpr size_t N_REASONS = 0 FOREACH_GC_REASON(COUNT_GC_REASON);
+#undef COUNT_GC_REASON
+};
+
+}  // namespace Gjs
+
+[[nodiscard]] const char* gjs_explain_gc_reason(JS::GCReason reason);
+
 #endif  // GJS_JSAPI_UTIL_H_
