@@ -1,26 +1,7 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
-/*
- * Copyright (c) 2013       Intel Corporation
- * Copyright (c) 2008-2010  litl, LLC
- *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to
- * deal in the Software without restriction, including without limitation the
- * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
- * sell copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
- * IN THE SOFTWARE.
- */
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2013 Intel Corporation
+// SPDX-FileCopyrightText: 2008-2010 litl, LLC
 
 #ifndef GI_FUNDAMENTAL_H_
 #define GI_FUNDAMENTAL_H_
@@ -32,7 +13,9 @@
 
 #include <js/TypeDecls.h>
 
+#include "gi/cwrapper.h"
 #include "gi/wrapperutils.h"
+#include "cjs/jsapi-util.h"  // for GjsAutoCallableInfo
 #include "cjs/macros.h"
 #include "util/log.h"
 
@@ -49,24 +32,20 @@ namespace JS { class CallArgs; }
 class FundamentalBase
     : public GIWrapperBase<FundamentalBase, FundamentalPrototype,
                            FundamentalInstance> {
+    friend class CWrapperPointerOps<FundamentalBase>;
     friend class GIWrapperBase<FundamentalBase, FundamentalPrototype,
                                FundamentalInstance>;
 
  protected:
     explicit FundamentalBase(FundamentalPrototype* proto = nullptr)
         : GIWrapperBase(proto) {}
-    ~FundamentalBase(void) {}
 
-    static const GjsDebugTopic debug_topic = GJS_DEBUG_GFUNDAMENTAL;
-    static constexpr const char* debug_tag = "fundamental";
+    static constexpr GjsDebugTopic DEBUG_TOPIC = GJS_DEBUG_GFUNDAMENTAL;
+    static constexpr const char* DEBUG_TAG = "fundamental";
 
     static const struct JSClassOps class_ops;
     static const struct JSClass klass;
 
-    // Helper methods
-
-    [[nodiscard]] const char* to_string_kind() const { return "fundamental"; }
-
     // Public API
 
  public:
@@ -86,7 +65,7 @@ class FundamentalPrototype
     GIObjectInfoUnrefFunction m_unref_function;
     GIObjectInfoGetValueFunction m_get_value_function;
     GIObjectInfoSetValueFunction m_set_value_function;
-    GICallableInfo* m_constructor_info;
+    GjsAutoCallableInfo m_constructor_info;
 
     explicit FundamentalPrototype(GIObjectInfo* info, GType gtype);
     ~FundamentalPrototype(void);
@@ -103,13 +82,31 @@ class FundamentalPrototype
         return m_constructor_info;
     }
 
-    void* call_ref_function(void* ptr) const { return m_ref_function(ptr); }
-    void call_unref_function(void* ptr) const { m_unref_function(ptr); }
-    [[nodiscard]] void* call_get_value_function(const GValue* value) const {
-        return m_get_value_function(value);
+    void* call_ref_function(void* ptr) const {
+        if (!m_ref_function)
+            return ptr;
+
+        return m_ref_function(ptr);
     }
-    void call_set_value_function(GValue* value, void* object) const {
-        m_set_value_function(value, object);
+    void call_unref_function(void* ptr) const {
+        if (m_unref_function)
+            m_unref_function(ptr);
+    }
+    [[nodiscard]] bool call_get_value_function(const GValue* value,
+                                               void** ptr_out) const {
+        if (!m_get_value_function)
+            return false;
+
+        *ptr_out = m_get_value_function(value);
+        return true;
+    }
+    bool call_set_value_function(GValue* value, void* object) const {
+        if (m_set_value_function) {
+            m_set_value_function(value, object);
+            return true;
+        }
+
+        return false;
     }
 
     // Helper methods
@@ -158,8 +155,8 @@ class FundamentalInstance
 
     void ref(void) { get_prototype()->call_ref_function(m_ptr); }
     void unref(void) { get_prototype()->call_unref_function(m_ptr); }
-    void set_value(GValue* gvalue) const {
-        get_prototype()->call_set_value_function(gvalue, m_ptr);
+    [[nodiscard]] bool set_value(GValue* gvalue) const {
+        return get_prototype()->call_set_value_function(gvalue, m_ptr);
     }
 
     GJS_JSAPI_RETURN_CONVENTION
@@ -176,8 +173,9 @@ class FundamentalInstance
     GJS_JSAPI_RETURN_CONVENTION
     static JSObject* object_for_c_ptr(JSContext* cx, void* gfundamental);
     GJS_JSAPI_RETURN_CONVENTION
-    static JSObject* object_for_gvalue(JSContext* cx, const GValue* gvalue,
-                                       GType gtype);
+    static bool object_for_gvalue(JSContext* cx, const GValue* gvalue,
+                                  GType gtype,
+                                  JS::MutableHandleObject object_out);
 
     static void* copy_ptr(JSContext* cx, GType gtype, void* gfundamental);
 };
