@@ -1,43 +1,20 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
-/*
- * Copyright (c) 2016 Philip Chimento
- *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to
- * deal in the Software without restriction, including without limitation the
- * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
- * sell copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
- * IN THE SOFTWARE.
- */
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2016 Philip Chimento
 
-#include "config.h"
+#include <locale.h>  // for setlocale, LC_ALL
+#include <stdint.h>
+#include <stdlib.h>  // for exit
+#include <string.h>
 
-#include <locale.h>
-#include <unistd.h>
-
-#include <glib.h>
-#include <glib/gstdio.h>
 #include <gio/gio.h>
+#include <girepository.h>
+#include <glib-object.h>
+#include <glib.h>
 
-#include "cjs/gjs.h"
-#include "cjs/mem.h"
+#include <cjs/gjs.h>
 
-G_GNUC_NORETURN
-static void
-bail_out(GjsContext *gjs_context,
-         const char *msg)
-{
+[[noreturn]] static void bail_out(GjsContext* gjs_context, const char* msg) {
     g_object_unref(gjs_context);
     g_print("Bail out! %s\n", msg);
     exit(1);
@@ -51,8 +28,6 @@ main(int argc, char **argv)
 
     /* The fact that this isn't the default is kind of lame... */
     g_setenv("GJS_DEBUG_OUTPUT", "stderr", false);
-    /* Jasmine library has some code style nits that trip this */
-    g_setenv("GJS_DISABLE_EXTRA_WARNINGS", "1", false);
 
     setlocale(LC_ALL, "");
 
@@ -60,13 +35,16 @@ main(int argc, char **argv)
         g_irepository_prepend_search_path(g_getenv("TOP_BUILDDIR"));
     } else {
         g_irepository_prepend_search_path(INSTTESTDIR);
-        g_irepository_prepend_library_path(PKGLIBDIR);
+        g_irepository_prepend_library_path(INSTTESTDIR);
     }
 
     const char *coverage_prefix = g_getenv("GJS_UNIT_COVERAGE_PREFIX");
     const char *coverage_output_path = g_getenv("GJS_UNIT_COVERAGE_OUTPUT");
     const char *search_path[] = { "resource:///org/gjs/jsunit", NULL };
 
+    if (coverage_prefix)
+        gjs_coverage_enable();
+
     GjsContext *cx = gjs_context_new_with_search_path((char **)search_path);
     GjsCoverage *coverage = NULL;
 
@@ -86,12 +64,19 @@ main(int argc, char **argv)
     bool success;
     int code;
 
-    success = gjs_context_eval(cx, "imports.minijasmine;", -1,
-                               "<jasmine>", &code, &error);
-    if (!success)
+    int exitcode_ignored;
+    if (!gjs_context_eval(cx, "imports.minijasmine;", -1, "<jasmine>",
+                          &exitcode_ignored, &error))
         bail_out(cx, error->message);
 
-    success = gjs_context_eval_file(cx, argv[1], &code, &error);
+    bool eval_as_module = argc >= 3 && strcmp(argv[2], "-m") == 0;
+    if (eval_as_module) {
+        uint8_t u8_exitcode_ignored;
+        success = gjs_context_eval_module_file(cx, argv[1],
+                                               &u8_exitcode_ignored, &error);
+    } else {
+        success = gjs_context_eval_file(cx, argv[1], &exitcode_ignored, &error);
+    }
     if (!success)
         bail_out(cx, error->message);
 
