@@ -1,37 +1,17 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
-/* Copyright 2010 litl, LLC.
- *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to
- * deal in the Software without restriction, including without limitation the
- * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
- * sell copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
- * IN THE SOFTWARE.
- */
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2010 litl, LLC.
 
 #include <config.h>
 
 #include <vector>
 
-#include <cairo-gobject.h>
 #include <cairo.h>
 #include <girepository.h>
 #include <glib.h>
 
 #include <js/Array.h>  // for JS::NewArrayObject
 #include <js/CallArgs.h>
-#include <js/Class.h>
 #include <js/Conversions.h>
 #include <js/PropertyDescriptor.h>  // for JSPROP_READONLY
 #include <js/PropertySpec.h>
@@ -39,24 +19,31 @@
 #include <js/TypeDecls.h>
 #include <js/Utility.h>  // for UniqueChars
 #include <js/Value.h>
+#include <js/ValueArray.h>
 #include <jsapi.h>  // for JS_SetElement
 
 #include "gi/arg-inl.h"
 #include "gi/arg.h"
 #include "gi/foreign.h"
-#include "cjs/jsapi-class.h"
+#include "cjs/enum-utils.h"
 #include "cjs/jsapi-util-args.h"
 #include "cjs/jsapi-util.h"
 #include "cjs/macros.h"
 #include "modules/cairo-private.h"
 
-#define _GJS_CAIRO_CONTEXT_DEFINE_FUNC_BEGIN(mname)              \
-    GJS_JSAPI_RETURN_CONVENTION                                  \
-    static bool mname##_func(JSContext* context, unsigned argc,  \
-                             JS::Value* vp) {                    \
-        GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr); \
-        if (!cr)                                                 \
-            return true;
+#define _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(cx, argc, vp, argv, obj) \
+    GJS_GET_THIS(cx, argc, vp, argv, obj);                              \
+    cairo_t* cr;                                                        \
+    if (!CairoContext::for_js_typecheck(cx, obj, &cr, &argv))           \
+        return false;                                                   \
+    if (!cr)                                                            \
+        return true;
+
+#define _GJS_CAIRO_CONTEXT_DEFINE_FUNC_BEGIN(mname)             \
+    GJS_JSAPI_RETURN_CONVENTION                                 \
+    static bool mname##_func(JSContext* context, unsigned argc, \
+                             JS::Value* vp) {                   \
+        _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj)
 
 #define _GJS_CAIRO_CONTEXT_DEFINE_FUNC_END                               \
     return gjs_cairo_check_status(context, cairo_status(cr), "context"); \
@@ -258,57 +245,37 @@ _GJS_CAIRO_CONTEXT_DEFINE_FUNC_BEGIN(method)                               \
     argv.rval().setUndefined();                                            \
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC_END
 
-[[nodiscard]] static JSObject* gjs_cairo_context_get_proto(JSContext*);
-
-GJS_DEFINE_PROTO_WITH_GTYPE("Context", cairo_context,
-                            CAIRO_GOBJECT_TYPE_CONTEXT,
-                            JSCLASS_BACKGROUND_FINALIZE)
-GJS_DEFINE_PRIV_FROM_JS(cairo_t, gjs_cairo_context_class);
-
-static void _gjs_cairo_context_construct_internal(JSObject* obj, cairo_t* cr) {
-    g_assert(!JS_GetPrivate(obj));
-    JS_SetPrivate(obj, cairo_reference(cr));
-}
-
-GJS_NATIVE_CONSTRUCTOR_DECLARE(cairo_context)
-{
-    GJS_NATIVE_CONSTRUCTOR_VARIABLES(cairo_context)
+GJS_JSAPI_RETURN_CONVENTION
+cairo_t* CairoContext::constructor_impl(JSContext* context,
+                                        const JS::CallArgs& argv) {
     cairo_t *cr;
 
-    GJS_NATIVE_CONSTRUCTOR_PRELUDE(cairo_context);
-
     JS::RootedObject surface_wrapper(context);
     if (!gjs_parse_call_args(context, "Context", argv, "o",
                              "surface", &surface_wrapper))
-        return false;
+        return nullptr;
 
-    cairo_surface_t* surface =
-        gjs_cairo_surface_get_surface(context, surface_wrapper);
+    cairo_surface_t* surface = CairoSurface::for_js(context, surface_wrapper);
     if (!surface)
-        return false;
+        return nullptr;
 
     cr = cairo_create(surface);
 
     if (!gjs_cairo_check_status(context, cairo_status(cr), "context"))
-        return false;
+        return nullptr;
 
-    _gjs_cairo_context_construct_internal(object, cr);
-    cairo_destroy(cr);
-
-    GJS_NATIVE_CONSTRUCTOR_FINISH(cairo_context);
-
-    return true;
+    return cr;
 }
 
-static void gjs_cairo_context_finalize(JSFreeOp*, JSObject* obj) {
-    using AutoCairoContext = GjsAutoPointer<cairo_t, cairo_t, cairo_destroy>;
-    AutoCairoContext cr = static_cast<cairo_t*>(JS_GetPrivate(obj));
-    JS_SetPrivate(obj, nullptr);
+void CairoContext::finalize_impl(JSFreeOp*, cairo_t* cr) {
+    if (!cr)
+        return;
+    cairo_destroy(cr);
 }
 
 /* Properties */
 // clang-format off
-JSPropertySpec gjs_cairo_context_proto_props[] = {
+const JSPropertySpec CairoContext::proto_props[] = {
     JS_STRING_SYM_PS(toStringTag, "Context", JSPROP_READONLY),
     JS_PS_END};
 // clang-format on
@@ -398,9 +365,7 @@ dispose_func(JSContext *context,
              unsigned   argc,
              JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, rec, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, rec, obj);
 
     cairo_destroy(cr);
     JS_SetPrivate(obj, nullptr);
@@ -415,9 +380,7 @@ appendPath_func(JSContext *context,
                 unsigned   argc,
                 JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::RootedObject path_wrapper(context);
 
@@ -425,8 +388,8 @@ appendPath_func(JSContext *context,
                              "path", &path_wrapper))
         return false;
 
-    cairo_path_t* path = gjs_cairo_path_get_path(context, path_wrapper);
-    if (!path)
+    cairo_path_t* path;
+    if (!CairoPath::for_js_typecheck(context, path_wrapper, &path, &argv))
         return false;
 
     cairo_append_path(cr, path);
@@ -440,9 +403,7 @@ copyPath_func(JSContext *context,
               unsigned   argc,
               JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     cairo_path_t *path;
 
@@ -450,7 +411,11 @@ copyPath_func(JSContext *context,
         return false;
 
     path = cairo_copy_path(cr);
-    argv.rval().setObjectOrNull(gjs_cairo_path_from_path(context, path));
+    JSObject* retval = CairoPath::take_c_ptr(context, path);
+    if (!retval)
+        return false;
+
+    argv.rval().setObject(*retval);
     return true;
 }
 
@@ -460,9 +425,7 @@ copyPathFlat_func(JSContext *context,
                   unsigned   argc,
                   JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     cairo_path_t *path;
 
@@ -470,7 +433,11 @@ copyPathFlat_func(JSContext *context,
         return false;
 
     path = cairo_copy_path_flat(cr);
-    argv.rval().setObjectOrNull(gjs_cairo_path_from_path(context, path));
+    JSObject* retval = CairoPath::take_c_ptr(context, path);
+    if (!retval)
+        return false;
+
+    argv.rval().setObject(*retval);
     return true;
 }
 
@@ -480,9 +447,7 @@ mask_func(JSContext *context,
           unsigned   argc,
           JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::RootedObject pattern_wrapper(context);
 
@@ -490,8 +455,7 @@ mask_func(JSContext *context,
                              "pattern", &pattern_wrapper))
         return false;
 
-    cairo_pattern_t* pattern =
-        gjs_cairo_pattern_get_pattern(context, pattern_wrapper);
+    cairo_pattern_t* pattern = CairoPattern::for_js(context, pattern_wrapper);
     if (!pattern)
         return false;
 
@@ -510,9 +474,7 @@ maskSurface_func(JSContext *context,
                  unsigned   argc,
                  JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::RootedObject surface_wrapper(context);
     double x, y;
@@ -523,8 +485,7 @@ maskSurface_func(JSContext *context,
                              "y", &y))
         return false;
 
-    cairo_surface_t* surface =
-        gjs_cairo_surface_get_surface(context, surface_wrapper);
+    cairo_surface_t* surface = CairoSurface::for_js(context, surface_wrapper);
     if (!surface)
         return false;
 
@@ -543,9 +504,7 @@ setDash_func(JSContext *context,
              unsigned   argc,
              JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     guint i;
     JS::RootedObject dashes(context);
@@ -604,9 +563,7 @@ setSource_func(JSContext *context,
                unsigned   argc,
                JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::RootedObject pattern_wrapper(context);
 
@@ -614,8 +571,7 @@ setSource_func(JSContext *context,
                              "pattern", &pattern_wrapper))
         return false;
 
-    cairo_pattern_t* pattern =
-        gjs_cairo_pattern_get_pattern(context, pattern_wrapper);
+    cairo_pattern_t* pattern = CairoPattern::for_js(context, pattern_wrapper);
     if (!pattern)
         return false;
 
@@ -635,9 +591,7 @@ setSourceSurface_func(JSContext *context,
                       unsigned   argc,
                       JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::RootedObject surface_wrapper(context);
     double x, y;
@@ -648,8 +602,7 @@ setSourceSurface_func(JSContext *context,
                              "y", &y))
         return false;
 
-    cairo_surface_t* surface =
-        gjs_cairo_surface_get_surface(context, surface_wrapper);
+    cairo_surface_t* surface = CairoSurface::for_js(context, surface_wrapper);
     if (!surface)
         return false;
 
@@ -669,9 +622,7 @@ showText_func(JSContext *context,
               unsigned   argc,
               JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::UniqueChars utf8;
 
@@ -695,9 +646,7 @@ selectFontFace_func(JSContext *context,
                     unsigned   argc,
                     JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, argv, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, argv, obj);
 
     JS::UniqueChars family;
     cairo_font_slant_t slant;
@@ -724,9 +673,7 @@ popGroup_func(JSContext *context,
               unsigned   argc,
               JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, rec, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, rec, obj);
 
     cairo_pattern_t *pattern;
     JSObject *pattern_wrapper;
@@ -757,9 +704,7 @@ getSource_func(JSContext *context,
                unsigned   argc,
                JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, rec, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, rec, obj);
 
     cairo_pattern_t *pattern;
     JSObject *pattern_wrapper;
@@ -791,12 +736,9 @@ getTarget_func(JSContext *context,
                unsigned   argc,
                JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, rec, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, rec, obj);
 
     cairo_surface_t *surface;
-    JSObject *surface_wrapper;
 
     if (argc > 0) {
         gjs_throw(context, "Context.getTarget() takes no arguments");
@@ -808,7 +750,7 @@ getTarget_func(JSContext *context,
         return false;
 
     /* surface belongs to the context, so keep the reference */
-    surface_wrapper = gjs_cairo_surface_from_surface(context, surface);
+    JSObject* surface_wrapper = CairoSurface::from_c_ptr(context, surface);
     if (!surface_wrapper) {
         /* exception already set */
         return false;
@@ -825,12 +767,9 @@ getGroupTarget_func(JSContext *context,
                     unsigned   argc,
                     JS::Value *vp)
 {
-    GJS_GET_PRIV(context, argc, vp, rec, obj, cairo_t, cr);
-    if (!cr)
-        return true;
+    _GJS_CAIRO_CONTEXT_GET_PRIV_CR_CHECKED(context, argc, vp, rec, obj);
 
     cairo_surface_t *surface;
-    JSObject *surface_wrapper;
 
     if (argc > 0) {
         gjs_throw(context, "Context.getGroupTarget() takes no arguments");
@@ -842,7 +781,7 @@ getGroupTarget_func(JSContext *context,
         return false;
 
     /* surface belongs to the context, so keep the reference */
-    surface_wrapper = gjs_cairo_surface_from_surface(context, surface);
+    JSObject* surface_wrapper = CairoSurface::from_c_ptr(context, surface);
     if (!surface_wrapper) {
         /* exception already set */
         return false;
@@ -853,7 +792,8 @@ getGroupTarget_func(JSContext *context,
     return true;
 }
 
-JSFunctionSpec gjs_cairo_context_proto_funcs[] = {
+// clang-format off
+const JSFunctionSpec CairoContext::proto_funcs[] = {
     JS_FN("$dispose", dispose_func, 0, 0),
     JS_FN("appendPath", appendPath_func, 0, 0),
     JS_FN("arc", arc_func, 0, 0),
@@ -953,37 +893,14 @@ JSFunctionSpec gjs_cairo_context_proto_funcs[] = {
     JS_FN("userToDevice", userToDevice_func, 0, 0),
     JS_FN("userToDeviceDistance", userToDeviceDistance_func, 0, 0),
     JS_FS_END};
-
-JSFunctionSpec gjs_cairo_context_static_funcs[] = { JS_FS_END };
-
-JSObject *
-gjs_cairo_context_from_context(JSContext *context,
-                               cairo_t *cr)
-{
-    JS::RootedObject proto(context, gjs_cairo_context_get_proto(context));
-    JS::RootedObject object(context,
-        JS_NewObjectWithGivenProto(context, &gjs_cairo_context_class, proto));
-    if (!object)
-        return nullptr;
-
-    _gjs_cairo_context_construct_internal(object, cr);
-
-    return object;
-}
-
-cairo_t *
-gjs_cairo_context_get_context(JSContext       *context,
-                              JS::HandleObject object)
-{
-    return priv_from_js(context, object);
-}
+// clang-format on
 
 [[nodiscard]] static bool context_to_g_argument(
     JSContext* context, JS::Value value, const char* arg_name,
-    GjsArgumentType argument_type, GITransfer transfer, bool may_be_null,
+    GjsArgumentType argument_type, GITransfer transfer, GjsArgumentFlags flags,
     GIArgument* arg) {
     if (value.isNull()) {
-        if (!may_be_null) {
+        if (!(flags & GjsArgumentFlags::MAY_BE_NULL)) {
             GjsAutoChar display_name =
                 gjs_argument_display_name(arg_name, argument_type);
             gjs_throw(context, "%s may not be null", display_name.get());
@@ -995,9 +912,7 @@ gjs_cairo_context_get_context(JSContext       *context,
     }
 
     JS::RootedObject obj(context, &value.toObject());
-    cairo_t *cr;
-
-    cr = gjs_cairo_context_get_context(context, obj);
+    cairo_t* cr = CairoContext::for_js(context, obj);
     if (!cr)
         return false;
     if (transfer == GI_TRANSFER_EVERYTHING)
@@ -1013,8 +928,8 @@ context_from_g_argument(JSContext             *context,
                         JS::MutableHandleValue value_p,
                         GIArgument            *arg)
 {
-    JSObject* obj =
-        gjs_cairo_context_from_context(context, gjs_arg_get<cairo_t*>(arg));
+    JSObject* obj = CairoContext::from_c_ptr(
+        context, static_cast<cairo_t*>(arg->v_pointer));
     if (!obj) {
         gjs_throw(context, "Could not create Cairo context");
         return false;
@@ -1031,12 +946,10 @@ static bool context_release_argument(JSContext*, GITransfer transfer,
     return true;
 }
 
-static GjsForeignInfo foreign_info = {
-    context_to_g_argument,
-    context_from_g_argument,
-    context_release_argument
-};
-
 void gjs_cairo_context_init(void) {
+    static GjsForeignInfo foreign_info = {context_to_g_argument,
+                                          context_from_g_argument,
+                                          context_release_argument};
+
     gjs_struct_foreign_register("cairo", "Context", &foreign_info);
 }
