@@ -1,23 +1,6 @@
 // application/javascript;version=1.8
-// Copyright 2013 Giovanni Campagna
-//
-// Permission is hereby granted, free of charge, to any person obtaining a copy
-// of this software and associated documentation files (the "Software"), to
-// deal in the Software without restriction, including without limitation the
-// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
-// sell copies of the Software, and to permit persons to whom the Software is
-// furnished to do so, subject to the following conditions:
-//
-// The above copyright notice and this permission notice shall be included in
-// all copies or substantial portions of the Software.
-//
-// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
-// IN THE SOFTWARE.
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2013 Giovanni Campagna
 
 const Legacy = imports._legacy;
 const {Gio, CjsPrivate, GObject} = imports.gi;
@@ -26,7 +9,6 @@ let Gtk;
 let BuilderScope;
 
 function _init() {
-
     Gtk = this;
 
     Gtk.children = GObject.__gtkChildren__;
@@ -44,13 +26,15 @@ function _init() {
     }
 
     Gtk.Widget.prototype._init = function (params) {
-        if (this.constructor[Gtk.template]) {
+        let wrapper = this;
+
+        if (wrapper.constructor[Gtk.template]) {
             if (!BuilderScope) {
-                Gtk.Widget.set_connect_func.call(this.constructor,
+                Gtk.Widget.set_connect_func.call(wrapper.constructor,
                     (builder, obj, signalName, handlerName, connectObj, flags) => {
                         const swapped = flags & GObject.ConnectFlags.SWAPPED;
                         const closure = _createClosure(
-                            builder, this, handlerName, swapped, connectObj);
+                            builder, wrapper, handlerName, swapped, connectObj);
 
                         if (flags & GObject.ConnectFlags.AFTER)
                             obj.connect_after(signalName, closure);
@@ -60,21 +44,23 @@ function _init() {
             }
         }
 
-        GObject.Object.prototype._init.call(this, params);
+        wrapper = GObject.Object.prototype._init.call(wrapper, params) ?? wrapper;
 
-        if (this.constructor[Gtk.template]) {
-            let children = this.constructor[Gtk.children] || [];
+        if (wrapper.constructor[Gtk.template]) {
+            let children = wrapper.constructor[Gtk.children] || [];
             for (let child of children) {
-                this[child.replace(/-/g, '_')] =
-                    this.get_template_child(this.constructor, child);
+                wrapper[child.replace(/-/g, '_')] =
+                    wrapper.get_template_child(wrapper.constructor, child);
             }
 
-            let internalChildren = this.constructor[Gtk.internalChildren] || [];
+            let internalChildren = wrapper.constructor[Gtk.internalChildren] || [];
             for (let child of internalChildren) {
-                this[`_${child.replace(/-/g, '_')}`] =
-                    this.get_template_child(this.constructor, child);
+                wrapper[`_${child.replace(/-/g, '_')}`] =
+                    wrapper.get_template_child(wrapper.constructor, child);
             }
         }
+
+        return wrapper;
     };
 
     Gtk.Widget._classInit = function (klass) {
